@@ -20,19 +20,33 @@
 #include <string.h>
 #include <libaudcore/audstrings.h>
 
-#include "audconfig.h"
-#include "chardet.h"
 #include "config.h"
+#include "debug.h"
 #include "i18n.h"
 #include "main.h"
-#include "debug.h"
+#include "misc.h"
 
 #ifdef USE_CHARDET
 #  include <libguess.h>
 #endif
 
-gchar *
-cd_str_to_utf8(const gchar * str)
+static gchar * cd_chardet_to_utf8 (const gchar * str, gssize len,
+ gsize * arg_bytes_read, gsize * arg_bytes_write, GError ** error);
+
+static gchar * str_to_utf8_fallback (const gchar * str)
+{
+    gchar * out = g_strconcat (str, _("  (invalid UTF-8)"), NULL);
+
+    for (gchar * c = out; * c; c ++)
+    {
+        if (* c & 0x80)
+            * c = '?';
+    }
+
+    return out;
+}
+
+static gchar * cd_str_to_utf8 (const gchar * str)
 {
     gchar *out_str;
 
@@ -74,21 +88,16 @@ cd_str_to_utf8(const gchar * str)
     if ((out_str = cd_chardet_to_utf8(str, strlen(str), NULL, NULL, NULL)) != NULL)
         return out_str;
 
-    /* assume encoding associated with locale */
-    if ((out_str = g_locale_to_utf8(str, -1, NULL, NULL, NULL)) != NULL)
-        return out_str;
-
     /* all else fails, we mask off character codes >= 128, replace with '?' */
     return str_to_utf8_fallback(str);
 }
 
-gchar *
-cd_chardet_to_utf8(const gchar * str, gssize len, gsize * arg_bytes_read,
-                   gsize * arg_bytes_write, GError ** error)
+static gchar * cd_chardet_to_utf8 (const gchar * str, gssize len,
+ gsize * arg_bytes_read, gsize * arg_bytes_write, GError ** error)
 {
-#ifdef USE_CHARDET
-    gchar *det = NULL, *encoding = NULL;
-#endif
+    if (error)
+        * error = NULL;
+
     gchar *ret = NULL;
     gsize *bytes_read, *bytes_write;
     gsize my_bytes_read, my_bytes_write;
@@ -118,31 +127,33 @@ cd_chardet_to_utf8(const gchar * str, gssize len, gsize * arg_bytes_read,
 
         return ret;
     }
+
 #ifdef USE_CHARDET
-    if (cfg.chardet_detector)
-        det = cfg.chardet_detector;
+    gchar * det = get_string (NULL, "chardet_detector");
 
-    if (det)
+    if (det[0])
     {
-	AUDDBG("guess encoding (%s) %s\n", det, str);
-        encoding = (gchar *) libguess_determine_encoding(str, len, det);
+        AUDDBG("guess encoding (%s) %s\n", det, str);
+        const gchar * encoding = libguess_determine_encoding (str, len, det);
         AUDDBG("encoding = %s\n", encoding);
-        if (encoding == NULL)
-            goto fallback;
-
-        ret = g_convert(str, len, "UTF-8", encoding, bytes_read, bytes_write, error);
+        if (encoding)
+            ret = g_convert (str, len, "UTF-8", encoding, bytes_read,
+             bytes_write, (error && * error) ? NULL : error);
     }
 
-fallback:
+    g_free (det);
 #endif
 
     /* If detection failed or was not enabled, try fallbacks (if there are any) */
-    if (ret == NULL && cfg.chardet_fallback_s != NULL)
+    if (! ret)
     {
-        gchar **enc;
-        for (enc = cfg.chardet_fallback_s; *enc != NULL; enc++)
+        gchar * fallbacks = get_string (NULL, "chardet_fallback");
+        gchar * * split = g_strsplit_set (fallbacks, " ,:;|/", -1);
+
+        for (gchar * * enc = split; * enc; enc ++)
         {
-            ret = g_convert(str, len, "UTF-8", *enc, bytes_read, bytes_write, error);
+            ret = g_convert (str, len, "UTF-8", * enc, bytes_read, bytes_write,
+             (error && * error) ? NULL : error);
             if (len == *bytes_read)
                 break;
             else {
@@ -150,11 +161,20 @@ fallback:
                 ret = NULL;
             }
         }
+
+        g_strfreev (split);
+        g_free (fallbacks);
     }
 
+    /* First fallback: locale (duh!) */
+    if (ret == NULL)
+        ret = g_locale_to_utf8 (str, len, bytes_read, bytes_write,
+         (error && * error) ? NULL : error);
+
     /* The final fallback is ISO-8859-1, if no other is specified or conversions fail */
     if (ret == NULL)
-        ret = g_convert(str, len, "UTF-8", "ISO-8859-1", bytes_read, bytes_write, error);
+        ret = g_convert (str, len, "UTF-8", "ISO-8859-1", bytes_read,
+         bytes_write, (error && * error) ? NULL : error);
 
     if (ret != NULL)
     {
@@ -171,9 +191,7 @@ fallback:
     return NULL; /* If we have no idea, return NULL. */
 }
 
-
-void chardet_init(void)
+void chardet_init (void)
 {
-    str_to_utf8 = cd_str_to_utf8;
-    chardet_to_utf8 = cd_chardet_to_utf8;
+    str_set_utf8_impl (cd_str_to_utf8, cd_chardet_to_utf8);
 }
