@@ -1,6 +1,6 @@
 /*
  * general.h
- * Copyright 2010 John Lindgren
+ * Copyright 2011 John Lindgren
  *
  * This file is part of Audacious.
  *
@@ -22,7 +22,16 @@
 #ifndef AUDACIOUS_GENERAL_H
 #define AUDACIOUS_GENERAL_H
 
+#include <glib.h>
+
+#include "plugins.h"
+
 void general_init (void);
 void general_cleanup (void);
 
+gboolean general_plugin_start (PluginHandle * plugin);
+void general_plugin_stop (PluginHandle * plugin);
+
+PluginHandle * general_plugin_by_widget (/* GtkWidget * */ void * widget);
+
 #endif
