@@ -19,8 +19,11 @@
 
 #include <gtk/gtk.h>
 
+#include <audacious/misc.h>
+
 void audgui_set_default_icon (void)
 {
+#ifndef _WIN32
     gtk_window_set_default_icon_name ("audacious");
+#endif
 }
-
