@@ -118,7 +118,7 @@ static gboolean skip_extended_header_3 (VFSFile * handle, gint * _size)
 
     size = GUINT32_FROM_BE (size);
 
-    AUDDBG ("Found v2.3 extended header, size = %d.\n", (gint) size);
+    TAGDBG ("Found v2.3 extended header, size = %d.\n", (gint) size);
 
     if (vfs_fseek (handle, size, SEEK_CUR))
         return FALSE;
@@ -136,7 +136,7 @@ static gboolean skip_extended_header_4 (VFSFile * handle, gint * _size)
 
     size = unsyncsafe32 (GUINT32_FROM_BE (size));
 
-    AUDDBG ("Found v2.4 extended header, size = %d.\n", (gint) size);
+    TAGDBG ("Found v2.4 extended header, size = %d.\n", (gint) size);
 
     if (vfs_fseek (handle, size - 4, SEEK_CUR))
         return FALSE;
@@ -155,17 +155,17 @@ static gboolean validate_header (ID3v2Header * header, gboolean is_footer)
 
     header->size = unsyncsafe32 (GUINT32_FROM_BE (header->size));
 
-    AUDDBG ("Found ID3v2 %s:\n", is_footer ? "footer" : "header");
-    AUDDBG (" magic = %.3s\n", header->magic);
-    AUDDBG (" version = %d\n", (gint) header->version);
-    AUDDBG (" revision = %d\n", (gint) header->revision);
-    AUDDBG (" flags = %x\n", (gint) header->flags);
-    AUDDBG (" size = %d\n", (gint) header->size);
+    TAGDBG ("Found ID3v2 %s:\n", is_footer ? "footer" : "header");
+    TAGDBG (" magic = %.3s\n", header->magic);
+    TAGDBG (" version = %d\n", (gint) header->version);
+    TAGDBG (" revision = %d\n", (gint) header->revision);
+    TAGDBG (" flags = %x\n", (gint) header->flags);
+    TAGDBG (" size = %d\n", (gint) header->size);
     return TRUE;
 }
 
 static gboolean read_header (VFSFile * handle, gint * version, gboolean *
- syncsafe, gsize * offset, gint * header_size, gint * data_size, gint *
+ syncsafe, gint64 * offset, gint * header_size, gint * data_size, gint *
  footer_size)
 {
     ID3v2Header header, footer;
@@ -203,7 +203,7 @@ static gboolean read_header (VFSFile * handle, gint * version, gboolean *
     }
     else
     {
-        gsize end = vfs_fsize (handle);
+        gint64 end = vfs_fsize (handle);
 
         if (end < 0)
             return FALSE;
@@ -256,7 +256,7 @@ static gboolean read_header (VFSFile * handle, gint * version, gboolean *
         * data_size -= extended_size;
     }
 
-    AUDDBG ("Offset = %d, header size = %d, data size = %d, footer size = "
+    TAGDBG ("Offset = %d, header size = %d, data size = %d, footer size = "
      "%d.\n", (gint) * offset, * header_size, * data_size, * footer_size);
 
     return TRUE;
@@ -303,17 +303,17 @@ static gboolean read_frame (VFSFile * handle, gint max_size, gint version,
     if (header.size > max_size || header.size == 0)
         return FALSE;
 
-    AUDDBG ("Found frame:\n");
-    AUDDBG (" key = %.4s\n", header.key);
-    AUDDBG (" size = %d\n", (gint) header.size);
-    AUDDBG (" flags = %x\n", (gint) header.flags);
+    TAGDBG ("Found frame:\n");
+    TAGDBG (" key = %.4s\n", header.key);
+    TAGDBG (" size = %d\n", (gint) header.size);
+    TAGDBG (" flags = %x\n", (gint) header.flags);
 
     * frame_size = sizeof (ID3v2FrameHeader) + header.size;
     sprintf (key, "%.4s", header.key);
 
     if (header.flags & (ID3_FRAME_COMPRESSED | ID3_FRAME_ENCRYPTED))
     {
-        AUDDBG ("Hit compressed/encrypted frame %s.\n", key);
+        TAGDBG ("Hit compressed/encrypted frame %s.\n", key);
         return FALSE;
     }
 
@@ -334,7 +334,7 @@ static gboolean read_frame (VFSFile * handle, gint max_size, gint version,
     if (syncsafe || (header.flags & ID3_FRAME_SYNCSAFE))
         * size = unsyncsafe (* data, * size);
 
-    AUDDBG ("Data size = %d.\n", * size);
+    TAGDBG ("Data size = %d.\n", * size);
     return TRUE;
 }
 
@@ -371,7 +371,7 @@ static void free_generic_frame (GenericFrame * frame)
 }
 
 static void read_all_frames (VFSFile * handle, gint version, gboolean syncsafe,
- gint data_size, mowgli_dictionary_t * dict)
+ gint data_size, mowgli_patricia_t * dict)
 {
     gint pos;
 
@@ -388,9 +388,9 @@ static void read_all_frames (VFSFile * handle, gint version, gboolean syncsafe,
 
         pos += frame_size;
 
-        if (mowgli_dictionary_retrieve (dict, key) != NULL)
+        if (mowgli_patricia_retrieve (dict, key) != NULL)
         {
-            AUDDBG ("Discarding duplicate frame %s.\n", key);
+            TAGDBG ("Discarding duplicate frame %s.\n", key);
             g_free (data);
             continue;
         }
@@ -400,14 +400,14 @@ static void read_all_frames (VFSFile * handle, gint version, gboolean syncsafe,
         frame->data = data;
         frame->size = size;
 
-        mowgli_dictionary_add (dict, key, frame);
+        mowgli_patricia_add (dict, key, frame);
     }
 }
 
 static gboolean write_frame (VFSFile * handle, GenericFrame * frame, gint *
  frame_size)
 {
-    AUDDBG ("Writing frame %s, size %d\n", frame->key, frame->size);
+    TAGDBG ("Writing frame %s, size %d\n", frame->key, frame->size);
 
     ID3v2FrameHeader header;
 
@@ -432,22 +432,22 @@ typedef struct {
     gint written_size;
 } WriteState;
 
-static gint write_frame_cb (mowgli_dictionary_elem_t * elem, void * user)
+static gint write_frame_cb (const gchar * key, void * data, void * user)
 {
     WriteState * state = user;
     gint size;
-    if (! write_frame (state->file, elem->data, & size))
+    if (! write_frame (state->file, data, & size))
         return -1;
     state->written_size += size;
     return 0;
 }
 
-static gint writeAllFramesToFile (VFSFile * fd, mowgli_dictionary_t * dict)
+static gint writeAllFramesToFile (VFSFile * fd, mowgli_patricia_t * dict)
 {
     WriteState state = {fd, 0};
-    mowgli_dictionary_foreach (dict, write_frame_cb, & state);
+    mowgli_patricia_foreach (dict, write_frame_cb, & state);
 
-    AUDDBG ("Total frame bytes written = %d.\n", state.written_size);
+    TAGDBG ("Total frame bytes written = %d.\n", state.written_size);
     return state.written_size;
 }
 
@@ -485,12 +485,15 @@ static void associate_string (Tuple * tuple, gint field, const gchar *
     gchar * text = decode_text_frame (data, size);
 
     if (text == NULL || ! text[0])
+    {
+        g_free (text);
         return;
+    }
 
     if (customfield != NULL)
-        AUDDBG ("Custom field %s = %s.\n", customfield, text);
+        TAGDBG ("Custom field %s = %s.\n", customfield, text);
     else
-        AUDDBG ("Field %i = %s.\n", field, text);
+        TAGDBG ("Field %i = %s.\n", field, text);
 
     tuple_associate_string (tuple, field, customfield, text);
     g_free (text);
@@ -501,13 +504,16 @@ static void associate_int (Tuple * tuple, gint field, const gchar *
 {
     gchar * text = decode_text_frame (data, size);
 
-    if (text == NULL || ! text[0])
+    if (text == NULL || atoi (text) < 1)
+    {
+        g_free (text);
         return;
+    }
 
     if (customfield != NULL)
-        AUDDBG ("Custom field %s = %s.\n", customfield, text);
+        TAGDBG ("Custom field %s = %s.\n", customfield, text);
     else
-        AUDDBG ("Field %i = %s.\n", field, text);
+        TAGDBG ("Field %i = %s.\n", field, text);
 
     tuple_associate_int (tuple, field, customfield, atoi (text));
     g_free (text);
@@ -545,10 +551,10 @@ static void decode_private_info (Tuple * tuple, const guchar * data, gint size)
             if (!memcmp(value, SECONDARY_CLASS_GAMES_SONG, 16))
                 tuple_associate_string (tuple, -1, "media-class", "Game Soundtrack");
         } else {
-            AUDDBG("Unrecognised tag %s (Windows Media) ignored\n", text);
+            TAGDBG("Unrecognised tag %s (Windows Media) ignored\n", text);
         }
     } else {
-        AUDDBG("Unable to decode private data, skipping: %s\n", text);
+        TAGDBG("Unable to decode private data, skipping: %s\n", text);
     }
 
 DONE:
@@ -562,7 +568,7 @@ static void decode_comment (Tuple * tuple, const guchar * data, gint size)
     if (! decode_comment_frame (data, size, & lang, & type, & value))
         return;
 
-    AUDDBG ("Comment: lang = %s, type = %s, value = %s.\n", lang, type, value);
+    TAGDBG ("Comment: lang = %s, type = %s, value = %s.\n", lang, type, value);
 
     if (! type[0]) /* blank type == actual comment */
         tuple_associate_string (tuple, FIELD_COMMENT, NULL, value);
@@ -585,7 +591,7 @@ static void decode_txxx (Tuple * tuple, const guchar * data, gint size)
         return;
 
     gchar * value = separator + 1;
-    AUDDBG ("TXXX: %s = %s.\n", text, value);
+    TAGDBG ("TXXX: %s = %s.\n", text, value);
     tuple_associate_string (tuple, -1, text, value);
 
     g_free (text);
@@ -611,7 +617,7 @@ static gboolean decode_rva2_block (const guchar * * _data, gint * _size, gint *
     data += 4;
     size -= 4;
 
-    AUDDBG ("RVA2 block: channel = %d, adjustment = %d/%d, peak bits = %d\n",
+    TAGDBG ("RVA2 block: channel = %d, adjustment = %d/%d, peak bits = %d\n",
      * channel, * adjustment, * adjustment_unit, peak_bits);
 
     if (peak_bits > 0 && peak_bits < sizeof (gint) * 8)
@@ -631,7 +637,7 @@ static gboolean decode_rva2_block (const guchar * * _data, gint * _size, gint *
         data += bytes;
         size -= count;
 
-        AUDDBG ("RVA2 block: peak = %d/%d\n", * peak, * peak_unit);
+        TAGDBG ("RVA2 block: peak = %d/%d\n", * peak, * peak_unit);
     }
     else
     {
@@ -654,7 +660,7 @@ static void decode_rva2 (Tuple * tuple, const guchar * data, gint size)
 
     domain = (const gchar *) data;
 
-    AUDDBG ("RVA2 domain: %s\n", domain);
+    TAGDBG ("RVA2 domain: %s\n", domain);
 
     size -= strlen (domain) + 1;
     data += strlen (domain) + 1;
@@ -718,25 +724,25 @@ static void decode_genre (Tuple * tuple, const guchar * data, gint size)
         numericgenre = atoi (text);
 
     if (numericgenre > 0)
-    {
-        tuple_associate_string(tuple, FIELD_GENRE, NULL, convert_numericgenre_to_text(numericgenre));
-        return;
-    }
-    tuple_associate_string(tuple, FIELD_GENRE, NULL, text);
+        tuple_associate_string (tuple, FIELD_GENRE, NULL,
+         convert_numericgenre_to_text (numericgenre));
+    else
+        tuple_associate_string (tuple, FIELD_GENRE, NULL, text);
+
     g_free (text);
     return;
 }
 
 static GenericFrame * add_generic_frame (gint id, gint size,
- mowgli_dictionary_t * dict)
+ mowgli_patricia_t * dict)
 {
-    GenericFrame * frame = mowgli_dictionary_retrieve (dict, id3_frames[id]);
+    GenericFrame * frame = mowgli_patricia_retrieve (dict, id3_frames[id]);
 
     if (frame == NULL)
     {
         frame = g_malloc (sizeof (GenericFrame));
         strcpy (frame->key, id3_frames[id]);
-        mowgli_dictionary_add (dict, frame->key, frame);
+        mowgli_patricia_add (dict, frame->key, frame);
     }
     else
         g_free (frame->data);
@@ -746,18 +752,18 @@ static GenericFrame * add_generic_frame (gint id, gint size,
     return frame;
 }
 
-static void remove_frame (gint id, mowgli_dictionary_t * dict)
+static void remove_frame (gint id, mowgli_patricia_t * dict)
 {
-    GenericFrame * frame = mowgli_dictionary_retrieve (dict, id3_frames[id]);
+    GenericFrame * frame = mowgli_patricia_retrieve (dict, id3_frames[id]);
     if (frame == NULL)
         return;
 
-    AUDDBG ("Deleting frame %s.\n", id3_frames[id]);
-    mowgli_dictionary_delete (dict, id3_frames[id]);
+    TAGDBG ("Deleting frame %s.\n", id3_frames[id]);
+    mowgli_patricia_delete (dict, id3_frames[id]);
     free_generic_frame (frame);
 }
 
-static void add_text_frame (gint id, const gchar * text, mowgli_dictionary_t *
+static void add_text_frame (gint id, const gchar * text, mowgli_patricia_t *
  dict)
 {
     if (text == NULL)
@@ -766,7 +772,7 @@ static void add_text_frame (gint id, const gchar * text, mowgli_dictionary_t *
         return;
     }
 
-    AUDDBG ("Adding text frame %s = %s.\n", id3_frames[id], text);
+    TAGDBG ("Adding text frame %s = %s.\n", id3_frames[id], text);
     gint length = strlen (text);
 
     GenericFrame * frame = add_generic_frame (id, length + 1, dict);
@@ -774,7 +780,7 @@ static void add_text_frame (gint id, const gchar * text, mowgli_dictionary_t *
     memcpy (frame->data + 1, text, length);
 }
 
-static void add_comment_frame (const gchar * text, mowgli_dictionary_t * dict)
+static void add_comment_frame (const gchar * text, mowgli_patricia_t * dict)
 {
     if (text == NULL)
     {
@@ -782,7 +788,7 @@ static void add_comment_frame (const gchar * text, mowgli_dictionary_t * dict)
         return;
     }
 
-    AUDDBG ("Adding comment frame = %s.\n", text);
+    TAGDBG ("Adding comment frame = %s.\n", text);
     gint length = strlen (text);
     GenericFrame * frame = add_generic_frame (ID3_COMMENT, length + 5, dict);
 
@@ -792,13 +798,13 @@ static void add_comment_frame (const gchar * text, mowgli_dictionary_t * dict)
 }
 
 static void add_frameFromTupleStr (const Tuple * tuple, gint field, gint
- id3_field, mowgli_dictionary_t * dict)
+ id3_field, mowgli_patricia_t * dict)
 {
     add_text_frame (id3_field, tuple_get_string (tuple, field, NULL), dict);
 }
 
 static void add_frameFromTupleInt (const Tuple * tuple, gint field, gint
- id3_field, mowgli_dictionary_t * dict)
+ id3_field, mowgli_patricia_t * dict)
 {
     if (tuple_get_value_type (tuple, field, NULL) != TUPLE_INT)
     {
@@ -815,7 +821,7 @@ static gboolean id3v24_can_handle_file (VFSFile * handle)
 {
     gint version, header_size, data_size, footer_size;
     gboolean syncsafe;
-    gsize offset;
+    gint64 offset;
 
     return read_header (handle, & version, & syncsafe, & offset, & header_size,
      & data_size, & footer_size);
@@ -825,7 +831,7 @@ static gboolean id3v24_read_tag (Tuple * tuple, VFSFile * handle)
 {
     gint version, header_size, data_size, footer_size;
     gboolean syncsafe;
-    gsize offset;
+    gint64 offset;
     gint pos;
 
     if (! read_header (handle, & version, & syncsafe, & offset, & header_size,
@@ -896,7 +902,7 @@ static gboolean id3v24_read_tag (Tuple * tuple, VFSFile * handle)
             decode_rva2 (tuple, data, size);
             break;
           default:
-            AUDDBG ("Ignoring unsupported ID3 frame %s.\n", key);
+            TAGDBG ("Ignoring unsupported ID3 frame %s.\n", key);
             break;
         }
 
@@ -925,7 +931,7 @@ static gboolean parse_apic (const guchar * _data, gint size, gchar * * mime,
     * image_data = g_memdup (after, data + size - after);
     * image_size = data + size - after;
 
-    AUDDBG ("APIC: mime = %s, type = %d, desc = %s, size = %d.\n", * mime,
+    TAGDBG ("APIC: mime = %s, type = %d, desc = %s, size = %d.\n", * mime,
      * type, * desc, * image_size);
     return TRUE;
 }
@@ -935,7 +941,7 @@ static gboolean id3v24_read_image (VFSFile * handle, void * * image_data, gint *
 {
     gint version, header_size, data_size, footer_size, parsed;
     gboolean syncsafe;
-    gsize offset;
+    gint64 offset;
     gboolean found = FALSE;
 
     if (! read_header (handle, & version, & syncsafe, & offset, & header_size,
@@ -977,23 +983,23 @@ static gboolean id3v24_read_image (VFSFile * handle, void * * image_data, gint *
     return found;
 }
 
-static void free_frame_cb (mowgli_dictionary_elem_t * element, void * unused)
+static void free_frame_cb (const gchar * key, void * data, void * unused)
 {
-    free_generic_frame (element->data);
+    free_generic_frame (data);
 }
 
 static gboolean id3v24_write_tag (const Tuple * tuple, VFSFile * f)
 {
     gint version, header_size, data_size, footer_size;
     gboolean syncsafe;
-    gsize offset;
+    gint64 offset;
 
     if (! read_header (f, & version, & syncsafe, & offset, & header_size,
      & data_size, & footer_size))
         return FALSE;
 
     //read all frames into generic frames;
-    mowgli_dictionary_t * dict = mowgli_dictionary_create (strcasecmp);
+    mowgli_patricia_t * dict = mowgli_patricia_create (NULL);
     read_all_frames (f, version, syncsafe, data_size, dict);
 
     //make the new frames from tuple and replace in the dictionary the old frames with the new ones
@@ -1008,33 +1014,33 @@ static gboolean id3v24_write_tag (const Tuple * tuple, VFSFile * f)
     if (! offset)
     {
         if (! cut_beginning_tag (f, header_size + data_size + footer_size))
-            goto ERROR;
+            goto ERR;
     }
     else
     {
         if (offset + header_size + data_size + footer_size != vfs_fsize (f))
-            goto ERROR;
+            goto ERR;
         if (vfs_ftruncate (f, offset))
-            goto ERROR;
+            goto ERR;
     }
 
     offset = vfs_fsize (f);
 
     if (offset < 0 || vfs_fseek (f, offset, SEEK_SET) || ! write_header (f, 0,
      FALSE))
-        goto ERROR;
+        goto ERR;
 
     data_size = writeAllFramesToFile (f, dict);
 
     if (! write_header (f, data_size, TRUE) || vfs_fseek (f, offset, SEEK_SET)
      || ! write_header (f, data_size, FALSE))
-        goto ERROR;
+        goto ERR;
 
-    mowgli_dictionary_destroy (dict, free_frame_cb, NULL);
+    mowgli_patricia_destroy (dict, free_frame_cb, NULL);
     return TRUE;
 
-ERROR:
-    mowgli_dictionary_destroy (dict, free_frame_cb, NULL);
+ERR:
+    mowgli_patricia_destroy (dict, free_frame_cb, NULL);
     return FALSE;
 }
 
